#include    <avr/io.h>
#include    <stdio.h>
#include    <util/delay.h>
#include    "lcd.h"

/* sample: lcd_cmd(LCD_DISP_OFF); */
#define LCD_DISP_OFF             0x08   /* display off                            */
#define LCD_DISP_ON              0x0C   /* display on, cursor off                 */
#define LCD_DISP_ON_BLINK        0x0D   /* display on, cursor off, blink char     */
#define LCD_DISP_ON_CURSOR       0x0E   /* display on, cursor on                  */
#define LCD_DISP_ON_CURSOR_BLINK 0x0F   /* display on, cursor on, blink char      */

/* move cursor/shift display */
#define LCD_MOVE_CURSOR_LEFT     0x10   /* move cursor left  (decrement)          */
#define LCD_MOVE_CURSOR_RIGHT    0x14   /* move cursor right (increment)          */
#define LCD_MOVE_DISP_LEFT       0x18   /* shift display left                     */
#define LCD_MOVE_DISP_RIGHT      0x1C   /* shift display right                    */

void wait_ms(int miliSec)
{
    _delay_loop_2( 1*(F_CPU/(1000/4)) * miliSec);    // 4 Zyklen warteschleife
}

//////////////////////////////////////////////////////////////////////////////
//    lcd_send(..) - sendet ein Byte an LCD im 4-Bit-Modus
//    nicht direkt aufrufen sondern lcd_cmd oder lcd_write nutzen
//////////////////////////////////////////////////////////////////////////////
void lcd_send(char data)
{
    // aktuelles RS ermitteln
    // RS liegt am Port D2
    char rs=PORTD; 
    rs&=4; // rs wird auf logisch 1 gesetzt
    
    // High-Teil senden
    char tmp=data;
    tmp&=0xf0; // Bits 0-3 werden auf 0 gesetzt
    tmp|=rs;   // RS wird "eingeodert" ;-)
    PORTD=tmp; 

    // Schreibsignal
    PORTD |= (1<<3);     //sbi(PORTD,3);
    PORTD &= ~(1<<3);    //cbi(PORTD,3);
    
    // Low-Teil senden
    tmp=data;
    tmp&=0x0f; // Bits 4-7 werden auf 0 gesetzt
    tmp*=16;   // mal 16 --> Bits werden um 4 Stellen nach links geschoben
    tmp|=rs;
    PORTD=tmp;
 
    // Schreibsignal
    PORTD |= (1<<3);    //sbi(PORTD,3);
    PORTD &= ~(1<<3);   //cbi(PORTD,3);

    // verarbeiten lassen
    wait_ms(1);
}

//////////////////////////////////////////////////////////////////////////////
//    lcd_cmd(..) - sendet ein Kommando an LCD
//////////////////////////////////////////////////////////////////////////////
void lcd_cmd(char cmd)
{
    PORTD &= ~(1<<2); // cbi(PORTD,2);  RS lschen = Kommando
    lcd_send(cmd);    // senden
}

//////////////////////////////////////////////////////////////////////////////
//    lcd_write(..) - sendet ein Zeichen (Daten) an LCD
//////////////////////////////////////////////////////////////////////////////
void lcd_write(char text)
{
    PORTD |= (1<<2);  // sbi(PORTD,2);  RS setzen = Daten
    lcd_send(text);   // senden
}

//////////////////////////////////////////////////////////////////////////////
//    lcd_write(..) - sendet eine Zeichenkette an LCD
//    Die Zeichenkette muss mit 0x00 abgeschlossen sein
//
//    char* bedeutet einen Zeiger auf eine Zeichenkette
//
//    durch die * - Notation wird automatisch der Abschluss mit 0x00 gemacht
//
//////////////////////////////////////////////////////////////////////////////
void lcd_write_s(char* pText)
{
    while(pText[0]!=0)
    {
        lcd_write(pText[0]);
        pText++;
    }
}

//////////////////////////////////////////////////////////////////////////////
//    lcd_write(..) - sendet eine Zeichenkette an LCD
//////////////////////////////////////////////////////////////////////////////
/*void lcd_write(char* pText, int count)
{
    while(count!=0)
    {
        lcd_write(pText[0]);
        pText++;
        count--;
    }
}*/

//////////////////////////////////////////////////////////////////////////////
//    lcd_home(..) - Cursor auf Position 1,1
//////////////////////////////////////////////////////////////////////////////
void lcd_home(void)
{
    lcd_cmd(0x02);
    wait_ms(2);            // warten
}

//////////////////////////////////////////////////////////////////////////////
//    lcd_clear(..) - lscht die Anzeige im LCD
//////////////////////////////////////////////////////////////////////////////
void lcd_clear(void)
{
    lcd_cmd(0x01);
    wait_ms(2);            // warten
}

//////////////////////////////////////////////////////////////////////////////
//    lcd_on(..) - schaltet das LCD an
//////////////////////////////////////////////////////////////////////////////
void lcd_on(void)
{
    lcd_cmd(0x0E);
}

//////////////////////////////////////////////////////////////////////////////
//    lcd_off(..) - schaltet das LCD aus
//////////////////////////////////////////////////////////////////////////////
void lcd_off(void)
{
    lcd_cmd(0x08);
}

//////////////////////////////////////////////////////////////////////////////
//    lcd_goto(..) - setzt die Cursorposition
//////////////////////////////////////////////////////////////////////////////
void lcd_goto(int row, int col)
{
    row--;             // Null-basierend
    row&=0x01;         // sicherheitshalber
    row*=0x40;         // Zeile nach Bit 6 bringen
    col--;             // Null-basierend
    col&=0x0f;         // sicherheitshalber
    char tmp=row|col;
    tmp|=0x80;         // Cursor setzen
    lcd_cmd(tmp);      // senden
}

//////////////////////////////////////////////////////////////////////////////
//    lcd_init(..) - Schaltet die Ports und Initialisiert das LCD
//////////////////////////////////////////////////////////////////////////////
void lcd_init(void)
{
    // Port D = Ausgang
    DDRD=0xff;
    PORTD=0;

    // warten bist LCD-Controller gebootet
    wait_ms(200);        

    // 4-BitModus einschalten
    PORTD=0x20;            

    // Schreibsignal
    PORTD |= (1<<3);   // sbi(PORTD,3);
    PORTD &= ~(1<<3);  // cbi(PORTD,3);
    wait_ms(5);        // Zeit zum Umschalten lassen
    
    // ab hier im 4-Bit-Modus
    lcd_cmd(0x28);        // Funktions-Set: 2 Zeilen, 5x7 Matrix, 4 Bit
    
    //lcd_off();
    lcd_cmd(0x06);        // Entry Mode
    lcd_on();
    lcd_clear();
}

//////////////////////////////////////////////////////////////////////////////
//    lcd_charDef(..) - initialisiert benutzerdefinierte Zeichen fr das LCD
//////////////////////////////////////////////////////////////////////////////
void lcd_charDef(void)
{
    // 1. Kommando CGRAM + Adresse setzen
    // 2. Punktmatrix-Daten laden
    unsigned char charSet[64]={    
                                0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xFF,
                                0x00,0x00,0x00,0x00,0x00,0x00,0xFF,0xFF,
                                0x00,0x00,0x00,0x00,0x00,0xFF,0xFF,0xFF,
                                0x00,0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF,
                                0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF,0xFF,
                                0x00,0x00,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
                                0x00,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
                                0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
                              };
    int i=0;
    for (i=0;i<64;i++)
    {
        lcd_cmd(0x40+i);              // cmd 0x40 + Adr
        lcd_write(charSet[i]);        // Pixel
    }
}

